==> chromedriver_autoinstaller_max-1.0.5-py3-none-any.whl: chromedriver_autoinstaller_max/utils.py <==
# coding: utf-8
"""
Helper functions for filename and URL generation.
"""

import json
import logging
import os
import re
import shutil
import subprocess
import sys
import xml.etree.ElementTree as elemTree
import zipfile
from io import BytesIO
import platform as pf

import requests
import warnings
from urllib3.exceptions import InsecureRequestWarning
warnings.simplefilter('ignore',InsecureRequestWarning)
import ssl
ssl._create_default_https_context = ssl._create_unverified_context

# the original author 
#__author__ = "Yeongbin Jo <iam.yeongbin.jo@gmail.com>"
# forked from yeongbin-jo/python-chromedriver-autoinstaller
__author__ = "Max Yao <weng.32002@gmail.com>"

from typing import AnyStr, Optional


def get_chromedriver_filename():
    """
    Returns the filename of the binary for the current platform.
    :return: Binary filename
    """
    if sys.platform.startswith("win"):
        return "chromedriver.exe"
    return "chromedriver"


def get_variable_separator():
    """
    Returns the environment variable separator for the current platform.
    :return: Environment variable separator
    """
    if sys.platform.startswith("win"):
        return ";"
    return ":"


def get_platform_architecture(chrome_version=None):
    platform = ""
    architecture = "64"
    if sys.platform.startswith("linux") and sys.maxsize > 2**32:
        platform = "linux"
        architecture = "64"
    elif sys.platform == "darwin":
        platform = "mac"
        # At some points, the release naming for Apple arm changed;
        # Looking in http://chromedriver.storage.googleapis.com/, the changeover happened across these releases:
        # 106.0.5249.61/chromedriver_mac_arm64.zip
        # 106.0.5249.21/chromedriver_mac64_m1.zip
        # Mac architecture naming changed again as of the transition to CfT
        # 115.0.5763.0/mac-arm64/chromedriver-mac-arm64.zip'
        # 115.0.5763.0/mac-x64/chromedriver-mac-x64.zip'
        if pf.processor() == "arm":
            if chrome_version is not None and chrome_version >= "115":
                print("CHROME >= 115, using mac-arm64 as architecture identifier")
                architecture = "-arm64"
            elif chrome_version is not None and chrome_version <= "106.0.5249.21":
                print("CHROME <= 106.0.5249.21, using mac64_m1 as architecture identifier")
                architecture = "64_m1"
            else:
                architecture = "_arm64"
        elif pf.processor() == "i386":
            if chrome_version is not None and chrome_version >= "115":
                print("CHROME >= 115, using mac-x64 as architecture identifier")
                architecture = "-x64"
            else:
                architecture = "64"
        else:
            raise RuntimeError("Could not determine Mac processor architecture.")
    elif sys.platform.startswith("win"):
        platform = "win"
        architecture = "32"
    else:
        raise RuntimeError(
            "Could not determine chromedriver download URL for this platform."
        )
    return platform, architecture


def get_chromedriver_url(chromedriver_version, download_options, no_ssl=False):
    """
    Generates the download URL for current platform , architecture and the given version.
    Supports Linux, MacOS and Windows.

    :param chromedriver_version: ChromeDriver version string
    :param no_ssl:               Whether to use the encryption protocol when downloading the chrome driver
    :return:                     String. Download URL for chromedriver
    """
    url = None
    platform, architecture = get_platform_architecture(chromedriver_version)
    if chromedriver_version >= "115" or chromedriver_version is None:  # new CfT ChromeDriver versions have their URLs published, so we already have a list of options
        for option in download_options:
            if option["platform"] == platform + architecture:
                        url = option['url']
    else:  # old ChromeDriver versions use the old urls
        base_url = "chromedriver.storage.googleapis.com/"
        base_url = "http://" + base_url if no_ssl else "https://" + base_url
        url = base_url + chromedriver_version + "/chromedriver_" + platform + architecture + ".zip"
    return url

def find_binary_in_path(filename):
    """
    Searches for a binary named `filename` in the current PATH. If an executable is found, its absolute path is returned
    else None.
    :param filename: Filename of the binary
    :return: Absolute path or None
    """
    if "PATH" not in os.environ:
        return None
    for directory in os.environ["PATH"].split(get_variable_separator()):
        binary = os.path.abspath(os.path.join(directory, filename))
        if os.path.isfile(binary) and os.access(binary, os.X_OK):
            return binary
    return None


def check_version(binary, required_version):
    try:
        version = subprocess.check_output([binary, "-v"])
        version = re.match(r".*?([\d.]+).*?", version.decode("utf-8"))[1]
        if version == required_version:
            return True
    except Exception:
        return False
    return False


def get_chrome_version():
    """
    :return: the version of chrome installed on client
    """
    version=None
    platform, _ = get_platform_architecture()
    if platform == "linux":
        path = get_linux_executable_path()
        with subprocess.Popen([path, "--version"], stdout=subprocess.PIPE) as proc:
            version = (
                proc.stdout.read()
                .decode("utf-8")
                .replace("Chromium", "")
                .replace("Google Chrome", "")
                .strip()
            )
    elif platform == "mac":
        process = subprocess.Popen(
            [
                "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
                "--version",
            ],
            stdout=subprocess.PIPE,
        )
        version = (
            process.communicate()[0]
            .decode("UTF-8")
            .replace("Google Chrome", "")
            .strip()
        )
    elif platform == "win":
        # check both of Program Files and Program Files (x86).
        # if the version isn't found on both of them, version is an empty string.
        chrome_dir_list = [
            "C:\\Program Files\\Google\\Chrome\\Application",
            "C:\\Program Files (x86)\\Google\\Chrome\\Application",
            "D:\\Program Files\\Google\\Chrome\\Application",
            "D:\\Program Files (x86)\\Google\\Chrome\\Application",
            "~\\AppData\\Local\\Google\\Chrome\\Application",
        ]
        for chrome_dir in chrome_dir_list:
            if chrome_dir[:1] == "~":
                chrome_dir = os.path.abspath(os.path.expanduser(chrome_dir))
            if os.path.exists(chrome_dir):
                dirs = [f.name for f in os.scandir(chrome_dir) if f.is_dir() and re.match("^[0-9.]+$", f.name)]
                version = max(dirs) if dirs else ''
    else:
        pass
    return version


def get_linux_executable_path():
    """
    Look through a list of candidates for Google Chrome executables that might
    exist, and return the full path to first one that does. Raise a ValueError
    if none do.

    :return: the full path to a Chrome executable on the system
    """
    for executable in (
        "google-chrome",
        "google-chrome-stable",
        "google-chrome-beta",
        "google-chrome-dev",
        "chromium-browser",
        "chromium",
    ):
        path = shutil.which(executable)
        if path is not None:
            return path
    raise ValueError("No chrome executable found on PATH")


def get_major_version(version):
    """
    :param version: the version of chrome
    :return: the major version of chrome
    """
    return version.split(".")[0]

def get_last_known_good_versions():
    download_options = None
    version_url = "googlechromelabs.github.io/chrome-for-testing/last-known-good-versions-with-downloads.json"
    version_url = f"http://{version_url}" if no_ssl else f"https://{version_url}"
    last_known_good_versions = None
    try:
        last_known_good_versions = json.loads(requests.get(version_url).text)
    except Exception as exc:
        print(exc)
        pass
    if not last_known_good_versions is None:
        if 'channels' in last_known_good_versions:
            if 'Stable' in last_known_good_versions['channels']:
                if 'downloads' in last_known_good_versions['channels']['Stable']:
                    if 'chromedriver' in last_known_good_versions['channels']['Stable']['downloads']:
                        download_options = last_known_good_versions['channels']['Stable']['downloads']['chromedriver']
    else:
        print("not able to get last know good versions")
    return download_options

def get_matched_chromedriver_version(chrome_version, no_ssl=False):
    """
    Try to find a specific version of ChromeDriver to match a version of cCrome

    :param chrome_version: the version of Chrome to match against
    :param no_ssl:         get version list using unsecured HTTP get
    :return:               String. The version of chromedriver that matches the Chrome version
                           None.   if no matching version of chromedriver was discovered
    """
    # Newer versions of chrome use the CfT publishing system
    browser_major_version = None
    download_options = None
    if chrome_version >= "115":
        browser_major_version = get_major_version(chrome_version)
        #version_url = "googlechromelabs.github.io/chrome-for-testing/known-good-versions-with-downloads.json"
        version_url = "googlechromelabs.github.io/chrome-for-testing/latest-versions-per-milestone-with-downloads.json"
        version_url = f"http://{version_url}" if no_ssl else f"https://{version_url}"
        latest_version_per_milestone_list = None
        try:
            latest_version_per_milestone_list = json.loads(requests.get(version_url).text)
        except Exception as exc:
            print(exc)
            pass
        # drill through object to make sure each key exists, to determin in driver download is available
        if not latest_version_per_milestone_list is None:
            if browser_major_version in latest_version_per_milestone_list['milestones']:
                milestone = latest_version_per_milestone_list['milestones'][browser_major_version]
                if 'downloads' in milestone:
                    if 'chromedriver' in milestone['downloads']:
                        download_options = milestone['downloads']['chromedriver']
        else:
            print("get latest version per milestone list fail...")

        # get last-known-good-versions
        if download_options is None:
            download_options = get_last_known_good_versions()

    # check old versions of chrome using the old system
    else:
        version_url = "chromedriver.storage.googleapis.com"
        version_url = "http://" + version_url if no_ssl else "https://" + version_url
        root = None
        try:
            root = elemTree.fromstring(requests.get(version_url).text)
        except Exception as exc:
            pass
        if not root is None:
            for k in root.iter("{http://doc.s3.amazonaws.com/2006-03-01}Key"):
                if k.text.find(get_major_version(chrome_version) + ".") == 0:
                    browser_major_version = k.text.split("/")[0]
        else:
            print("not able to get tree element")

    return browser_major_version, download_options


def get_chromedriver_path():
    """
    :return: path of the chromedriver binary
    """
    return os.path.abspath(os.path.dirname(__file__))


def print_chromedriver_path():
    """
    Print the path of the chromedriver binary.
    """
    print(get_chromedriver_path())

def get_chromedriver_dir(path, major_version, make_version_dir: bool = True):
    chromedriver_dir = ""
    if path:
        if not os.path.isdir(path):
            raise ValueError(f"Invalid path: {path}")
        if make_version_dir:
            chromedriver_dir = os.path.join(os.path.abspath(path), major_version)
        else:
            chromedriver_dir = os.path.abspath(path)
    else:
        chromedriver_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), major_version)
    return chromedriver_dir

def download_chromedriver(path: Optional[AnyStr] = None, no_ssl: bool = False, make_version_dir: bool = True, detect_installed_version: bool = True):
    """
    Downloads, unzips and installs chromedriver.
    If a chromedriver binary is found in PATH it will be copied, otherwise downloaded.

    :param str path: Path of the directory where to save the downloaded chromedriver to.
    :param bool no_ssl: Determines whether or not to use the encryption protocol when downloading the chrome driver.
    :return: The file path of chromedriver
    """
    chromedriver_version = None
    chromedriver_dir = ""
    download_options = None

    chrome_version = None
    if detect_installed_version:
        chrome_version = get_chrome_version()

    if chrome_version is None:
        logging.debug("Chrome is not installed or not able to access")
        download_options = get_last_known_good_versions()
        chromedriver_dir = os.path.abspath(path)
    else:
        chromedriver_version, download_options = get_matched_chromedriver_version(chrome_version, no_ssl)
        if not chromedriver_version:
            logging.warning(
                "Can not find chromedriver for currently installed chrome version."
            )
            return
        major_version = get_major_version(chromedriver_version)
        chromedriver_dir = get_chromedriver_dir(path, major_version, make_version_dir)
    
    chromedriver_filename = get_chromedriver_filename()
    chromedriver_filepath = os.path.join(chromedriver_dir, chromedriver_filename)
    if not os.path.isfile(chromedriver_filepath) or not check_version(chromedriver_filepath, chromedriver_version):
        logging.info(f"Downloading chromedriver ({chromedriver_version})...")
        if not os.path.isdir(chromedriver_dir):
            os.makedirs(chromedriver_dir)
        url = get_chromedriver_url(chromedriver_version=chromedriver_version, download_options=download_options, no_ssl=no_ssl)
        response = None
        
        retry_idx_max = 5
        response_ok = False
        for retry_idx in range(retry_idx_max):
            #print("download webdriver try No:", retry_idx+1)
            if retry_idx>0:
                url = url.replace("https://","http://")
            print("download webdriver from url:", url)
            runtime_error_message = ""
            try:
                response = requests.get(url, verify=False)
                if not response is None:
                    if response.status_code != 200:
                        runtime_error_message = "URL Status Not OK: {url}"
                    else:
                        response_ok = True
            except Exception as exc:
                response = None
                runtime_error_message = "Failed to download chromedriver archive: {url}"

            if response_ok:
                break
        if not response_ok:
            raise RuntimeError(runtime_error_message)


        if not response is None:
            archive = BytesIO(response.content)
            with zipfile.ZipFile(archive) as zip_file:
                # v115+ have files in subdirectories- need to adjust filename before extracting
                for zip_info in zip_file.infolist():
                    if os.path.basename(zip_info.filename) == chromedriver_filename:
                        zip_info.filename = chromedriver_filename
                        zip_file.extract(zip_info, chromedriver_dir)
                        break
    else:
        logging.info("Chromedriver is already installed.")
    if not os.access(chromedriver_filepath, os.X_OK):
        os.chmod(chromedriver_filepath, 0o744)
    return chromedriver_filepath


if __name__ == "__main__":
    print(get_chrome_version())
    print(download_chromedriver(no_ssl=False))

==> chromedriver_autoinstaller_max-1.0.6-py3-none-any.whl: chromedriver_autoinstaller_max/utils.py <==
# coding: utf-8
"""
Helper functions for filename and URL generation.
"""

import json
import logging
import os
import re
import shutil
import subprocess
import sys
import xml.etree.ElementTree as elemTree
import zipfile
from io import BytesIO
import platform as pf

import requests
import warnings
from urllib3.exceptions import InsecureRequestWarning
warnings.simplefilter('ignore',InsecureRequestWarning)
import ssl
ssl._create_default_https_context = ssl._create_unverified_context

# the original author 
#__author__ = "Yeongbin Jo <iam.yeongbin.jo@gmail.com>"
# forked from yeongbin-jo/python-chromedriver-autoinstaller
__author__ = "Max Yao <weng.32002@gmail.com>"

from typing import AnyStr, Optional


def get_chromedriver_filename():
    """
    Returns the filename of the binary for the current platform.
    :return: Binary filename
    """
    if sys.platform.startswith("win"):
        return "chromedriver.exe"
    return "chromedriver"


def get_variable_separator():
    """
    Returns the environment variable separator for the current platform.
    :return: Environment variable separator
    """
    if sys.platform.startswith("win"):
        return ";"
    return ":"


def get_platform_architecture(chrome_version=None):
    platform = ""
    architecture = "64"
    if sys.platform.startswith("linux") and sys.maxsize > 2**32:
        platform = "linux"
        architecture = "64"
    elif sys.platform == "darwin":
        platform = "mac"
        # At some points, the release naming for Apple arm changed;
        # Looking in http://chromedriver.storage.googleapis.com/, the changeover happened across these releases:
        # 106.0.5249.61/chromedriver_mac_arm64.zip
        # 106.0.5249.21/chromedriver_mac64_m1.zip
        # Mac architecture naming changed again as of the transition to CfT
        # 115.0.5763.0/mac-arm64/chromedriver-mac-arm64.zip'
        # 115.0.5763.0/mac-x64/chromedriver-mac-x64.zip'
        if pf.processor() == "arm":
            if chrome_version is not None and chrome_version >= "115":
                print("CHROME >= 115, using mac-arm64 as architecture identifier")
                architecture = "-arm64"
            elif chrome_version is not None and chrome_version <= "106.0.5249.21":
                print("CHROME <= 106.0.5249.21, using mac64_m1 as architecture identifier")
                architecture = "64_m1"
            else:
                architecture = "_arm64"
        elif pf.processor() == "i386":
            if chrome_version is not None and chrome_version >= "115":
                print("CHROME >= 115, using mac-x64 as architecture identifier")
                architecture = "-x64"
            else:
                architecture = "64"
        else:
            raise RuntimeError("Could not determine Mac processor architecture.")
    elif sys.platform.startswith("win"):
        platform = "win"
        architecture = "32"
    else:
        raise RuntimeError(
            "Could not determine chromedriver download URL for this platform."
        )
    return platform, architecture


def get_chromedriver_url(chromedriver_version, download_options, no_ssl=False):
    """
    Generates the download URL for current platform , architecture and the given version.
    Supports Linux, MacOS and Windows.

    :param chromedriver_version: ChromeDriver version string
    :param no_ssl:               Whether to use the encryption protocol when downloading the chrome driver
    :return:                     String. Download URL for chromedriver
    """
    url = None
    platform, architecture = get_platform_architecture(chromedriver_version)
    if chromedriver_version >= "115" or chromedriver_version is None:  # new CfT ChromeDriver versions have their URLs published, so we already have a list of options
        for option in download_options:
            if option["platform"] == platform + architecture:
                        url = option['url']
    else:  # old ChromeDriver versions use the old urls
        base_url = "chromedriver.storage.googleapis.com/"
        base_url = "http://" + base_url if no_ssl else "https://" + base_url
        url = base_url + chromedriver_version + "/chromedriver_" + platform + architecture + ".zip"
    return url

def find_binary_in_path(filename):
    """
    Searches for a binary named `filename` in the current PATH. If an executable is found, its absolute path is returned
    else None.
    :param filename: Filename of the binary
    :return: Absolute path or None
    """
    if "PATH" not in os.environ:
        return None
    for directory in os.environ["PATH"].split(get_variable_separator()):
        binary = os.path.abspath(os.path.join(directory, filename))
        if os.path.isfile(binary) and os.access(binary, os.X_OK):
            return binary
    return None


def check_version(binary, required_version):
    try:
        version = subprocess.check_output([binary, "-v"])
        version = re.match(r".*?([\d.]+).*?", version.decode("utf-8"))[1]
        if version == required_version:
            return True
    except Exception:
        return False
    return False


def get_chrome_version():
    """
    :return: the version of chrome installed on client
    """
    version=None
    platform, _ = get_platform_architecture()
    if platform == "linux":
        path = get_linux_executable_path()
        with subprocess.Popen([path, "--version"], stdout=subprocess.PIPE) as proc:
            version = (
                proc.stdout.read()
                .decode("utf-8")
                .replace("Chromium", "")
                .replace("Google Chrome", "")
                .strip()
            )
    elif platform == "mac":
        process = subprocess.Popen(
            [
                "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
                "--version",
            ],
            stdout=subprocess.PIPE,
        )
        version = (
            process.communicate()[0]
            .decode("UTF-8")
            .replace("Google Chrome", "")
            .strip()
        )
    elif platform == "win":
        PROGRAMFILES = f"{os.environ.get('PROGRAMW6432') or os.environ.get('PROGRAMFILES')}\\Google\\Chrome\\Application"
        PROGRAMFILESX86 = f"{os.environ.get('PROGRAMFILES(X86)')}\\Google\\Chrome\\Application"
        
        path = PROGRAMFILES if os.path.exists(PROGRAMFILES) else PROGRAMFILESX86 if os.path.exists(PROGRAMFILESX86) else None

        dirs = [f.name for f in os.scandir(path) if f.is_dir() and re.match("^[0-9.]+$", f.name)] if path else None

        version = max(dirs) if dirs else None        
    else:
        pass
    return version


def get_linux_executable_path():
    """
    Look through a list of candidates for Google Chrome executables that might
    exist, and return the full path to first one that does. Raise a ValueError
    if none do.

    :return: the full path to a Chrome executable on the system
    """
    for executable in (
        "google-chrome",
        "google-chrome-stable",
        "google-chrome-beta",
        "google-chrome-dev",
        "chromium-browser",
        "chromium",
    ):
        path = shutil.which(executable)
        if path is not None:
            return path
    raise ValueError("No chrome executable found on PATH")


def get_major_version(version):
    """
    :param version: the version of chrome
    :return: the major version of chrome
    """
    return version.split(".")[0]

def get_last_known_good_versions(no_ssl=False):
    download_options = None
    version_url = "googlechromelabs.github.io/chrome-for-testing/last-known-good-versions-with-downloads.json"
    version_url = f"http://{version_url}" if no_ssl else f"https://{version_url}"
    last_known_good_versions = None
    try:
        last_known_good_versions = json.loads(requests.get(version_url).text)
    except Exception as exc:
        print(exc)
        pass
    if not last_known_good_versions is None:
        if 'channels' in last_known_good_versions:
            if 'Stable' in last_known_good_versions['channels']:
                if 'downloads' in last_known_good_versions['channels']['Stable']:
                    if 'chromedriver' in last_known_good_versions['channels']['Stable']['downloads']:
                        download_options = last_known_good_versions['channels']['Stable']['downloads']['chromedriver']
    else:
        print("not able to get last know good versions")
    return download_options

def get_matched_chromedriver_version(chrome_version, no_ssl=False):
    """
    Try to find a specific version of ChromeDriver to match a version of cCrome

    :param chrome_version: the version of Chrome to match against
    :param no_ssl:         get version list using unsecured HTTP get
    :return:               String. The version of chromedriver that matches the Chrome version
                           None.   if no matching version of chromedriver was discovered
    """
    # Newer versions of chrome use the CfT publishing system
    browser_major_version = None
    download_options = None
    if chrome_version >= "115":
        browser_major_version = get_major_version(chrome_version)
        #version_url = "googlechromelabs.github.io/chrome-for-testing/known-good-versions-with-downloads.json"
        version_url = "googlechromelabs.github.io/chrome-for-testing/latest-versions-per-milestone-with-downloads.json"
        version_url = f"http://{version_url}" if no_ssl else f"https://{version_url}"
        latest_version_per_milestone_list = None
        try:
            latest_version_per_milestone_list = json.loads(requests.get(version_url).text)
        except Exception as exc:
            print(exc)
            pass
        # drill through object to make sure each key exists, to determin in driver download is available
        if not latest_version_per_milestone_list is None:
            if browser_major_version in latest_version_per_milestone_list['milestones']:
                milestone = latest_version_per_milestone_list['milestones'][browser_major_version]
                if 'downloads' in milestone:
                    if 'chromedriver' in milestone['downloads']:
                        download_options = milestone['downloads']['chromedriver']
        else:
            print("get latest version per milestone list fail...")

        # get last-known-good-versions
        if download_options is None:
            download_options = get_last_known_good_versions(no_ssl)

    # check old versions of chrome using the old system
    else:
        version_url = "chromedriver.storage.googleapis.com"
        version_url = "http://" + version_url if no_ssl else "https://" + version_url
        root = None
        try:
            root = elemTree.fromstring(requests.get(version_url).text)
        except Exception as exc:
            pass
        if not root is None:
            for k in root.iter("{http://doc.s3.amazonaws.com/2006-03-01}Key"):
                if k.text.find(get_major_version(chrome_version) + ".") == 0:
                    browser_major_version = k.text.split("/")[0]
        else:
            print("not able to get tree element")

    return browser_major_version, download_options


def get_chromedriver_path():
    """
    :return: path of the chromedriver binary
    """
    return os.path.abspath(os.path.dirname(__file__))


def print_chromedriver_path():
    """
    Print the path of the chromedriver binary.
    """
    print(get_chromedriver_path())

def get_chromedriver_dir(path, major_version, make_version_dir: bool = True):
    chromedriver_dir = ""
    if path:
        if not os.path.isdir(path):
            raise ValueError(f"Invalid path: {path}")
        if make_version_dir:
            chromedriver_dir = os.path.join(os.path.abspath(path), major_version)
        else:
            chromedriver_dir = os.path.abspath(path)
    else:
        chromedriver_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), major_version)
    return chromedriver_dir

def download_chromedriver(path: Optional[AnyStr] = None, no_ssl: bool = False, make_version_dir: bool = True, detect_installed_version: bool = True):
    """
    Downloads, unzips and installs chromedriver.
    If a chromedriver binary is found in PATH it will be copied, otherwise downloaded.

    :param str path: Path of the directory where to save the downloaded chromedriver to.
    :param bool no_ssl: Determines whether or not to use the encryption protocol when downloading the chrome driver.
    :return: The file path of chromedriver
    """
    chromedriver_version = None
    chromedriver_dir = ""
    download_options = None

    chrome_version = None
    if detect_installed_version:
        chrome_version = get_chrome_version()

    if chrome_version is None:
        logging.debug("Chrome is not installed or not able to access")
        download_options = get_last_known_good_versions(no_ssl)
        chromedriver_dir = os.path.abspath(path)
    else:
        chromedriver_version, download_options = get_matched_chromedriver_version(chrome_version, no_ssl)
        if not chromedriver_version:
            logging.warning(
                "Can not find chromedriver for currently installed chrome version."
            )
            return
        major_version = get_major_version(chromedriver_version)
        chromedriver_dir = get_chromedriver_dir(path, major_version, make_version_dir)
    
    chromedriver_filename = get_chromedriver_filename()
    chromedriver_filepath = os.path.join(chromedriver_dir, chromedriver_filename)
    if not os.path.isfile(chromedriver_filepath) or not check_version(chromedriver_filepath, chromedriver_version):
        logging.info(f"Downloading chromedriver ({chromedriver_version})...")
        if not os.path.isdir(chromedriver_dir):
            os.makedirs(chromedriver_dir)
        url = get_chromedriver_url(chromedriver_version=chromedriver_version, download_options=download_options, no_ssl=no_ssl)
        response = None
        
        retry_idx_max = 5
        response_ok = False
        for retry_idx in range(retry_idx_max):
            #print("download webdriver try No:", retry_idx+1)
            if retry_idx>0:
                url = url.replace("https://","http://")
            print("download webdriver from url:", url)
            runtime_error_message = ""
            try:
                response = requests.get(url, verify=False)
                if not response is None:
                    if response.status_code != 200:
                        runtime_error_message = "URL Status Not OK: {url}"
                    else:
                        response_ok = True
            except Exception as exc:
                response = None
                runtime_error_message = "Failed to download chromedriver archive: {url}"

            if response_ok:
                break
        if not response_ok:
            raise RuntimeError(runtime_error_message)


        if not response is None:
            archive = BytesIO(response.content)
            with zipfile.ZipFile(archive) as zip_file:
                # v115+ have files in subdirectories- need to adjust filename before extracting
                for zip_info in zip_file.infolist():
                    if os.path.basename(zip_info.filename) == chromedriver_filename:
                        zip_info.filename = chromedriver_filename
                        zip_file.extract(zip_info, chromedriver_dir)
                        break
    else:
        logging.info("Chromedriver is already installed.")
    if not os.access(chromedriver_filepath, os.X_OK):
        os.chmod(chromedriver_filepath, 0o744)
    return chromedriver_filepath


if __name__ == "__main__":
    print(get_chrome_version())
    print(download_chromedriver(no_ssl=False))
